/*--------------------------------*-C-*---------------------------------*
 * File:	xpm.c
 *----------------------------------------------------------------------*
 * $Id: xpm.c,v 1.37 2000/07/16 09:04:37 gcw Exp $
 *
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 1997      Carsten Haitzler <raster@zip.com.au>
 * Copyright (C) 1997,1998 Oezguer Kesim <kesim@math.fu-berlin.de>
 * Copyright (C) 1998,1999 Geoff Wing <gcw@pobox.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *---------------------------------------------------------------------*/

#include "../config.h"		/* NECESSARY */
#include "rxvt.h"		/* NECESSARY */
#include "xpm.intpro"		/* PROTOS for internal routines */

#ifdef XPM_BACKGROUND

/*
 * These GEOM strings indicate absolute size/position:
 * @ `WxH+X+Y'
 * @ `WxH+X'    -> Y = X
 * @ `WxH'      -> Y = X = 50
 * @ `W+X+Y'    -> H = W
 * @ `W+X'      -> H = W, Y = X
 * @ `W'        -> H = W, X = Y = 50
 * @ `0xH'      -> H *= H/100, X = Y = 50 (W unchanged)
 * @ `Wx0'      -> W *= W/100, X = Y = 50 (H unchanged)
 * @ `=+X+Y'    -> (H, W unchanged)
 * @ `=+X'      -> Y = X (H, W unchanged)
 *
 * These GEOM strings adjust position relative to current position:
 * @ `+X+Y'
 * @ `+X'       -> Y = X
 *
 * And this GEOM string is for querying current scale/position:
 * @ `?'
 */
/* EXTPROTO */
int
rxvt_scale_pixmap(rxvt_t *r, const char *geom)
{
    int             w = 0, h = 0, x = 0, y = 0;
    int             n, flags, changed = 0;
    char           *p, *str;

#define MAXLEN_GEOM		sizeof("[1000x1000+1000+1000]")

    if (geom == NULL)
	return 0;
    str = MALLOC(MAXLEN_GEOM + 1);
    if (!STRCMP(geom, "?")) {
	sprintf(str, "[%dx%d+%d+%d]",	/* can't presume snprintf? */
		(r->h->bgPixmap.w < 9999 ? r->h->bgPixmap.w : 9999),
		(r->h->bgPixmap.h < 9999 ? r->h->bgPixmap.h : 9999),
		(r->h->bgPixmap.x < 9999 ? r->h->bgPixmap.x : 9999),
		(r->h->bgPixmap.y < 9999 ? r->h->bgPixmap.y : 9999));
	rxvt_xterm_seq(r, XTerm_title, str, CHAR_ST);
	FREE(str);
	return 0;
    }

    if ((p = STRCHR(geom, ';')) == NULL)
	p = STRCHR(geom, '\0');
    n = (p - geom);
    if (n <= MAXLEN_GEOM) {
	STRNCPY(str, geom, n);
	str[n] = '\0';

	flags = XParseGeometry(str, &x, &y, (unsigned int *)&w,
			       (unsigned int *)&h);
	if (!flags) {
	    flags |= WidthValue;
	    w = 0;
	}			/* default is tile */
	if (flags & WidthValue) {
	    if (!(flags & XValue))
		x = 50;

	    if (!(flags & HeightValue))
		h = w;

	    if (w && !h) {
		w = r->h->bgPixmap.w * ((float)w / 100);
		h = r->h->bgPixmap.h;
	    } else if (h && !w) {
		w = r->h->bgPixmap.w;
		h = r->h->bgPixmap.h * ((float)h / 100);
	    }
	    if (w > 1000)
		w = 1000;
	    if (h > 1000)
		h = 1000;

	    if (r->h->bgPixmap.w != w) {
		r->h->bgPixmap.w = w;

		changed++;
	    }
	    if (r->h->bgPixmap.h != h) {
		r->h->bgPixmap.h = h;

		changed++;
	    }
	}
	if (!(flags & YValue)) {
	    if (flags & XNegative)
		flags |= YNegative;
	    y = x;
	}

	if (!(flags & WidthValue) && geom[0] != '=') {
	    x += r->h->bgPixmap.x;
	    y += r->h->bgPixmap.y;
	} else {
	    if (flags & XNegative)
		x += 100;
	    if (flags & YNegative)
		y += 100;
	}
	MIN_IT(x, 100);
	MIN_IT(y, 100);
	MAX_IT(x, 0);
	MAX_IT(y, 0);
	if (r->h->bgPixmap.x != x) {
	    r->h->bgPixmap.x = x;

	    changed++;
	}
	if (r->h->bgPixmap.y != y) {
	    r->h->bgPixmap.y = y;

	    changed++;
	}
    }
    FREE(str);
    return changed;
}

/* EXTPROTO */
void
rxvt_resize_pixmap(rxvt_t *r)
{
    XGCValues       gcvalue;
    GC              gc;
    int             tiled = 0;
    unsigned int    width = TermWin_TotalWidth();
    unsigned int    height = TermWin_TotalHeight();

    if (r->TermWin.pixmap != None)
	XFreePixmap(r->Xdisplay, r->TermWin.pixmap);

    if (r->h->bgPixmap.pixmap == None) {	/* So be it: I'm not using pixmaps */
	r->TermWin.pixmap = None;
	if (!(r->Options & Opt_transparent) || r->h->am_transparent == 0)
	    XSetWindowBackground(r->Xdisplay, r->TermWin.vt,
				 r->PixColors[Color_bg]);
	return;
    }

    gcvalue.foreground = r->PixColors[Color_bg];
    gc = XCreateGC(r->Xdisplay, r->TermWin.vt, GCForeground, &gcvalue);

    if (r->h->bgPixmap.pixmap != None) {	/* we have a specified pixmap */
	int             w = r->h->bgPixmap.w, h = r->h->bgPixmap.h,
			x = r->h->bgPixmap.x, y = r->h->bgPixmap.y;

	/*
	 * don't zoom pixmap too much nor expand really small pixmaps
	 */
	if (w > 1000 || h > 1000)
	    w = 1;
	else if (width > (10 * r->h->xpmAttr.width)
		 || height > (10 * r->h->xpmAttr.height))
	    w = 0;		/* tile */

	if (w == 0) {
	    /* basic X tiling - let the X server do it */
	    r->TermWin.pixmap = XCreatePixmap(r->Xdisplay, r->TermWin.vt,
					      r->h->xpmAttr.width,
					      r->h->xpmAttr.height, r->Xdepth);
	    XCopyArea(r->Xdisplay, r->h->bgPixmap.pixmap, r->TermWin.pixmap, gc,
		      0, 0, r->h->xpmAttr.width, r->h->xpmAttr.height, 0, 0);

	    tiled = 1;
	} else {
	    float           p, incr;
	    Pixmap          tmp;

	    r->TermWin.pixmap = XCreatePixmap(r->Xdisplay, r->TermWin.vt,
					      width, height, r->Xdepth);
	    /*
	     * horizontal scaling
	     */
	    incr = (float)r->h->xpmAttr.width;
	    p = 0;

	    if (w == 1) {	/* display one image, no horizontal scaling */
		incr = width;
		if (r->h->xpmAttr.width <= width) {
		    w = r->h->xpmAttr.width;
		    x = x * (width - w) / 100;	/* beware! order */
		    w += x;
		} else {
		    x = 0;
		    w = width;
		}
	    } else if (w < 10) {	/* fit W images across screen */
		incr *= w;
		x = 0;
		w = width;
	    } else {
		incr *= 100.0 / w;
		if (w < 100) {	/* contract */
		    w = (w * width) / 100;
		    if (x >= 0) {	/* position */
			float           pos;

			pos = (float)x / 100 * width - (w / 2);

			x = (width - w);
			if (pos <= 0)
			    x = 0;
			else if (pos < x)
			    x = pos;
		    } else {
			x = x * (width - w) / 100;	/* beware! order */
		    }
		    w += x;
		} else {	/* expand */
		    if (x > 0) {	/* position */
			float           pos;

			pos = (float)x / 100 * r->h->xpmAttr.width - (incr / 2);
			p = r->h->xpmAttr.width - (incr);
			if (pos <= 0)
			    p = 0;
			else if (pos < p)
			    p = pos;
		    }
		    x = 0;
		    w = width;
		}
	    }
	    incr /= width;

	    tmp = XCreatePixmap(r->Xdisplay, r->TermWin.vt,
				width, r->h->xpmAttr.height, r->Xdepth);
	    XFillRectangle(r->Xdisplay, tmp, gc, 0, 0, width,
			   r->h->xpmAttr.height);

	    for ( /*nil */ ; x < w; x++, p += incr) {
		if (p >= r->h->xpmAttr.width)
		    p = 0;
		if (x < 0)
		    continue;
		/* copy one column from the original pixmap to the tmp pixmap */
		XCopyArea(r->Xdisplay, r->h->bgPixmap.pixmap, tmp, gc,
			  (int)p, 0, 1, r->h->xpmAttr.height, x, 0);
	    }

	    /*
	     * vertical scaling
	     */
	    incr = (float)r->h->xpmAttr.height;
	    p = 0;

	    if (h == 1) {	/* display one image, no vertical scaling */
		incr = height;
		if (r->h->xpmAttr.height <= height) {
		    h = r->h->xpmAttr.height;
		    y = y * (height - h) / 100;	/* beware! order */
		    h += y;
		} else {
		    y = 0;
		    h = height;
		}
	    } else if (h < 10) {	/* fit H images across screen */
		incr *= h;
		y = 0;
		h = height;
	    } else {
		incr *= 100.0 / h;
		if (h < 100) {	/* contract */
		    h = (h * height) / 100;
		    if (y >= 0) {	/* position */
			float           pos;

			pos = (float)y / 100 * height - (h / 2);

			y = (height - h);
			if (pos < 0.0f)
			    y = 0;
			else if (pos < y)
			    y = pos;
		    } else {
			y = y * (height - h) / 100;	/* beware! order */
		    }
		    h += y;
		} else {	/* expand */
		    if (y > 0) {	/* position */
			float           pos;

			pos = (float)y / 100 * r->h->xpmAttr.height - (incr / 2);
			p = r->h->xpmAttr.height - (incr);
			if (pos < 0)
			    p = 0;
			else if (pos < p)
			    p = pos;
		    }
		    y = 0;
		    h = height;
		}
	    }
	    incr /= height;

	    if (y > 0)
		XFillRectangle(r->Xdisplay, r->TermWin.pixmap, gc, 0, 0, width,
			       y);
	    if (h < height)
		XFillRectangle(r->Xdisplay, r->TermWin.pixmap, gc, 0, h, width,
			       height - h + 1);
	    for ( /*nil */ ; y < h; y++, p += incr) {
		if (p >= r->h->xpmAttr.height)
		    p = 0;
		if (y < 0)
		    continue;
		/* copy one row from the tmp pixmap to the main pixmap */
		XCopyArea(r->Xdisplay, tmp, r->TermWin.pixmap, gc,
			  0, (int)p, width, 1, 0, y);
	    }
	    XFreePixmap(r->Xdisplay, tmp);
	}
    }
    XSetWindowBackgroundPixmap(r->Xdisplay, r->TermWin.vt, r->TermWin.pixmap);
    XFreeGC(r->Xdisplay, gc);
    r->h->am_transparent = 0;

    XClearWindow(r->Xdisplay, r->TermWin.vt);

    XSync(r->Xdisplay, False);
}

/* EXTPROTO */
Pixmap
rxvt_set_bgPixmap(rxvt_t *r, const char *file)
{
    char           *f;

    assert(file != NULL);

    if (r->h->bgPixmap.pixmap != None) {
	XFreePixmap(r->Xdisplay, r->h->bgPixmap.pixmap);
	r->h->bgPixmap.pixmap = None;
    }
    XSetWindowBackground(r->Xdisplay, r->TermWin.vt, r->PixColors[Color_bg]);

    if (*file != '\0') {
/*      XWindowAttributes attr; */

	/*
	 * we already have the required attributes
	 */
/*      XGetWindowAttributes(r->Xdisplay, r->TermWin.vt, &attr); */

	r->h->xpmAttr.closeness = 30000;
	r->h->xpmAttr.colormap = r->Xcmap;
	r->h->xpmAttr.visual = r->h->Xvisual;
	r->h->xpmAttr.depth = r->Xdepth;
	r->h->xpmAttr.valuemask = (XpmCloseness | XpmColormap | XpmVisual |
				   XpmDepth | XpmSize | XpmReturnPixels);

	/* search environment variables here too */
	f = (char *)rxvt_File_find(file, ".xpm", r->h->rs[Rs_path]);
	if (f == NULL
	    || XpmReadFileToPixmap(r->Xdisplay, Xroot, f,
				   &r->h->bgPixmap.pixmap, NULL,
				   &r->h->xpmAttr)) {
	    char           *p;

	    /* semi-colon delimited */
	    if ((p = STRCHR(file, ';')) == NULL)
		p = STRCHR(file, '\0');

	    rxvt_print_error("couldn't load XPM file \"%.*s\"", (p - file),
			     file);
	}
	FREE(f);
    }
    rxvt_resize_pixmap(r);
    return r->h->bgPixmap.pixmap;
}

#endif				/* XPM_BACKGROUND */
